PG_ANALISE = """
<|layout|columns=.68fr auto 1fr|class_name=header_container|

<|part|class_name=header_logo|
<|Cashd|text|height=30px|width=30px|>
|>

<|part|class_name=align_item_stretch|
<|{nav_analise_val}|toggle|lov={nav_analise_lov}|on_change={lambda s: s.elem_analise.update_content(s, nav_analise_val[0])}|>
|>

<|part|class_name=text_right|class_name=header_top_right_corner|
<|🗕|button|on_action=btn_mudar_minimizado|>
<|🗖|button|on_action=btn_mudar_maximizado|>
<|✖|button|on_action=btn_encerrar|>
|>

|>

<|part|partial={elem_analise}|class_name=container|>
"""


ELEM_PLOT = """
<|layout|columns=1fr 1fr 2fr|class_name=top_controls|


<|{dropdown_tipo_val}|selector|lov={dropdown_tipo_lov}|dropdown|>

<|{dropdown_periodo_val}|selector|lov={dropdown_periodo_lov}|dropdown|>

<|{slider_val}|slider|lov={slider_lov}|text_anchor=botom|>

|>

<center>
<|Atualizar|button|on_action={btn_gerar_main_plot}|>
</center>

<br />

<|chart|figure={main_plot}|height=360px|>
"""

ELEM_HIST = """
<br />

<|layout|columns=1.2fr .7fr|

<|part|
<|layout|columns=.9 .1fr|class_name=interactive_header|
# Últimas transações

<|↻|button|on_action={btn_atualizar_df_ult_transac}|>
|>
<|{df_ult_transac}|table|paginated|page_size=10|page_size_options={[25, 50]}|height=360px|>
|>

<|part|
# Maiores saldos

<|{df_maiores_saldos}|table|show_all|height=412px|>
|>

|>
"""



from configparser import ConfigParser
from tempfile import TemporaryFile, TemporaryDirectory
import pytest
import sqlite3
import os

from cashd.backup import (
    conf,
    DB_FILE,
    CONFIG_FILE,
    BACKUP_PATH,
    read_db_size,
    write_current_size,
    parse_list_from_config,
    parse_list_to_config,
    copy_file,
    rename_on_db_folder,
    check_sqlite,
    read_last_recorded_size,
    write_add_backup_place,
    write_rm_backup_place,
    load,
    run,
)


SCRIPT_PATH = os.path.split(os.path.realpath(__file__))[0]


def test_dbsize_read():
    with TemporaryFile(delete=False) as data_file, TemporaryFile(
        delete=False
    ) as config_file:
        data_file.write(b"testando um teste testoso...")
        write_current_size(
            config_file=config_file.name,
            current_size=read_db_size(file_path=data_file.name),
        )
        config = ConfigParser()
        config.read(config_file.name)
        read_dbsize = int(config["file_sizes"]["dbsize"])
        assert read_dbsize == read_db_size(file_path=data_file.name)


def test_list_parse():
    string = "[\n\tc:/some/path,\n\tc:\\some\\other\\path]"
    expected_list = [r"c:/some/path", r"c:\some\other\path"]

    parsed_from = parse_list_from_config(string)
    parsed_to = parse_list_to_config(expected_list)
    assert parsed_from == expected_list
    assert parsed_to == string


def test_copy_file():
    tempfile = TemporaryFile(delete=False)
    tempfile.close()
    with TemporaryDirectory() as tempdir:
        # test success
        copy_file(tempfile.name, tempdir)
        assert len(os.listdir(tempdir)) == 1
        # test fail
        try:
            copy_file("thisisnotevenapath", tempdir, True)
        except FileNotFoundError:
            assert 1 == 1
    os.unlink(tempfile.name)


def test_rename_on_db_folder():
    db_folder = os.path.split(DB_FILE)[0]

    # test closed file
    file = TemporaryFile(delete=False, dir=db_folder)
    file.close()
    old_filename = os.path.split(file.name)[1]
    new_filename = "renamed.file"
    rename_on_db_folder(old_filename, new_filename)
    new_filepath = os.path.join(db_folder, new_filename)
    assert os.path.isfile(new_filepath)
    os.unlink(new_filepath)

    # test open file
    with TemporaryFile(dir=db_folder) as file:
        old_filename = os.path.split(file.name)[1]
        new_filename = "renamed.file"
        rename_on_db_folder(old_filename, new_filename)
        new_filepath = os.path.join(db_folder, new_filename)
        assert os.path.isfile(new_filepath)

    # test invalid file
    try:
        rename_on_db_folder("not a file", "go brrrrr", _raise=True)
    except Exception:
        assert True == True
    else:
        raise AssertionError("Should raise an error with an invalid file")


def test_check_sqlite():
    with TemporaryDirectory() as tempdir:
        # Create a valid SQLite database file
        valid_db_file = os.path.join(tempdir, "valid_database.db")
        con = sqlite3.connect(valid_db_file)
        cursor = con.cursor()
        cursor.execute("CREATE TABLE users (id INTEGER PRIMARY KEY, name TEXT)")
        con.commit()
        con.close()

        # Test valid database
        assert check_sqlite(valid_db_file) is True

        # Test invalid database
        invalid_db_file = os.path.join(tempdir, "invalid_database.txt")
        with open(invalid_db_file, "wt"):
            assert check_sqlite(invalid_db_file) is False

        # Test exception when _raise=True
        try:
            check_sqlite("oi", _raise=True)
        except Exception:
            pass
        else:
            raise AssertionError(
                "Expected FileExistsError, but no exception was raised"
            )


def test_read_db_size():
    # Create a temporary file
    with TemporaryFile(delete=False) as temp_file:
        temp_file.write(b"something")
        temp_file_path = temp_file.name

    # Test the function with an existing file
    size_existing = read_db_size(file_path=temp_file_path)
    assert size_existing == os.path.getsize(temp_file_path)

    # Test the function with a nonexistent file
    size_nonexistent = read_db_size(file_path="nonexistent_file.txt")
    assert size_nonexistent is None

    # Clean up the temporary file
    os.remove(temp_file_path)


def test_read_last_recorded_size_existing_section():
    # test existing section
    with TemporaryFile(delete=False) as temp_config_file:
        temp_config_file.write(b"[file_sizes]\ndbsize = 100\n")
        temp_config_file_path = temp_config_file.name

    size_existing = read_last_recorded_size(config_file=temp_config_file_path)
    assert size_existing == 100
    os.remove(temp_config_file_path)

    # test missing section
    with TemporaryFile(delete=False) as temp_config_file:
        temp_config_file.write(b"[other_section]\nkey = value\n")
        temp_config_file_path = temp_config_file.name

    size_nonexistent = read_last_recorded_size(config_file=temp_config_file_path)
    assert size_nonexistent == 0
    os.remove(temp_config_file_path)


def test_add_rm_backup_place():
    conf.read(CONFIG_FILE, "utf-8")
    current = parse_list_from_config(conf["default"]["backup_places"])
    path_to_add = r"C:\some\path\for\testing"

    # test add path to config
    write_add_backup_place(path_to_add)
    expected_add = current + [path_to_add]
    current_add = parse_list_from_config(conf["default"]["backup_places"])
    assert current_add == expected_add

    write_add_backup_place(path_to_add)
    current_add_repeated = parse_list_from_config(conf["default"]["backup_places"])
    # shall not add a path that already exists
    assert current_add_repeated == expected_add

    # test remove path from config
    added_idx = current_add.index(path_to_add)
    write_rm_backup_place(added_idx)
    current_rm = parse_list_from_config(conf["default"]["backup_places"])
    assert current_rm == current

    write_rm_backup_place("not even an index")
    # shall only perform action with valid indexes
    current_rm_invalid = parse_list_from_config(conf["default"]["backup_places"])
    assert current_rm_invalid == current_rm


def test_load():
    # test invalid file
    try:
        load(file="notfile", _raise=True)
    except OSError:
        assert True == True
    else:
        raise AssertionError("Should raise OSError")

    # test valid file
    dbdir = os.path.split(DB_FILE)[0]
    prev_stash = [f for f in os.listdir(dbdir) if "stash" in f]
    load(DB_FILE, _raise=True)
    new_stash = [f for f in os.listdir(dbdir) if "stash" in f]
    assert len(prev_stash) == len(new_stash) - 1

    # cleanup
    stashed = [f for f in new_stash if f not in prev_stash][0]
    os.remove(os.path.join(dbdir, stashed))


def test_run():
    # save current `backup_places`
    conf.read(CONFIG_FILE, "utf-8")
    prev_backup_places = parse_list_from_config(conf["default"]["backup_places"])

    # remove `backup_places` from config file
    curr_backup_places = parse_list_from_config(conf["default"]["backup_places"])
    while len(curr_backup_places) > 0:
        conf.read(CONFIG_FILE, "utf-8")
        write_rm_backup_place(0)
        curr_backup_places = parse_list_from_config(conf["default"]["backup_places"])

    # get current list of backups
    prev_saved_backups = os.listdir(BACKUP_PATH)

    # test invalid path
    write_add_backup_place("notpath")
    try:
        run(force=True, _raise=True)
    except NotADirectoryError:
        assert True == True
    else:
        raise AssertionError("Expected a NotADirectoryError")
    finally:
        write_rm_backup_place(0)

    # test valid path
    with TemporaryDirectory() as tempdir:
        write_add_backup_place(tempdir)
        try:
            run(force=False, _raise=True)
            run(force=True, _raise=True)
        except Exception as err:
            raise AssertionError(f"Expected no exception, got {type(err)}: {str(err)}")
        finally:
            write_rm_backup_place(0)

    # restore `backup_places` and cleanup
    for path in prev_backup_places:
        write_add_backup_place(path)
    for file in [f for f in os.listdir(BACKUP_PATH) if f not in prev_saved_backups]:
        os.remove(os.path.join(BACKUP_PATH, file))



from cashd.app import start_cashd
import sys


def window():
    if len(sys.argv) == 1:
        start_cashd(with_webview=True)
        return

    if "window" in sys.argv:
        start_cashd(with_webview=True)
        return

    if "browser" in sys.argv:
        start_cashd(with_webview=False)
        return


if __name__ == "__main__":
    start_cashd(with_webview=True)



# chamado em:
#   contas.ELEMENTO_REGS
SELECIONAR_CLIENTE_ETAPA = """
<center>
    <|{SLC_USUARIO}|selector|lov={NOMES_USUARIOS}|filter|propagate|height=520px|width=650px|on_change=chg_cliente_selecionado|>
</center>
"""

SELECIONAR_TRANSAC_ETAPA = """
<center>
    <|{SLC_TRANSAC}|selector|lov={TRANSACS_USUARIO}|filter|paginated|height=360px|>
</center>
"""

FORM_EDITAR_CLIENTE = """
<|layout|columns=1 1|columns[mobile]=1 1|class_name=container
__Primeiro Nome__*

__Sobrenome__*

<|{form_conta_selec.PrimeiroNome}|input|>

<|{form_conta_selec.Sobrenome}|input|>

__Apelido__

__Telefone__

<|{form_conta_selec.Apelido}|input|>

<|{form_conta_selec.Telefone}|input|label=DDD + 9 dígitos|>

__Endereço__

__Bairro__

<|{form_conta_selec.Endereco}|input|>

<|{form_conta_selec.Bairro}|input|>

__Cidade__*

__Estado__*

<|{form_conta_selec.Cidade}|input|>

<|{form_conta_selec.Estado}|input|>

_(*) Obrigatório_
<|{str(form_contas)}|text|>
"""

CONFIRMAR_TRANSAC = """
Valor: *<|{form_transac_selec.Valor}|>*

Data: *<|{form_transac_selec.DataTransac}|>*
"""


CONFIRMAR_CONTA = """
<|layout|columns=1 1|columns[mobile]=1 1|class_name=container
Primeiro Nome: *<|{form_conta_selec.PrimeiroNome}|>*

Sobrenome: *<|{form_conta_selec.Sobrenome}|>*

Apelido: *<|{form_conta_selec.Apelido}|>*

Telefone: *<|{form_conta_selec.Telefone}|>*

Endereço: *<|{form_conta_selec.Endereco}|>*

Bairro: *<|{form_conta_selec.Bairro}|>*

Cidade: *<|{form_conta_selec.Cidade}|>*

Estado: *<|{form_conta_selec.Estado}|>*
|>
"""



from cashd.db import DB_ENGINE
from cashd.prefs import BackupPrefsHandler

import sqlite3
from os import path, makedirs, rename
from datetime import datetime
import shutil
import configparser
import logging


####################
# GLOBAL VARS
####################

SCRIPT_PATH = path.split(path.realpath(__file__))[0]
BACKUP_PATH = path.join(SCRIPT_PATH, "backup")
CONFIG_FILE = path.join(SCRIPT_PATH, "configs", "backup.ini")
LOG_FILE = path.join(SCRIPT_PATH, "logs", "backup.log")
DB_FILE = path.join(SCRIPT_PATH, "data", "database.db")

settings = BackupPrefsHandler()

logger = logging.getLogger("cashd.backup")
logger.setLevel(logging.DEBUG)
logger.propagate = False

log_fmt = logging.Formatter("%(asctime)s :: %(levelname)s %(message)s")
log_handler = logging.FileHandler(LOG_FILE)
log_handler.setLevel(logging.DEBUG)
log_handler.setFormatter(log_fmt)

logger.addHandler(log_handler)


####################
# UTILS
####################


def copy_file(source_path, target_dir, _raise: bool = False):
    logger.debug("function call: copy_file")
    now = datetime.now().strftime("%Y-%m-%d_%H-%M-%S_%f")
    try:
        filename = f"backup_{now}.db"
        shutil.copyfile(source_path, path.join(target_dir, filename))
        logger.info(f"Copia de '{source_path}' criada em '{target_dir}'")
    except FileNotFoundError as xpt:
        logger.error(f"Erro realizando copia: {xpt}.", exc_info=1)
        if _raise:
            raise xpt


def rename_on_db_folder(current: str, new: str, _raise: bool = False):
    """
    Renomeia um arquivo na mesma pasta em que `DB_FILE` se encontra, se a
    operacao falhar porque o arquivo esta em uso, faz uma copia com o novo
    nome em vez de renomear.

    Levanta o ultimo erro que recebeu se ambas as operacoes falharem.
    """
    logger.debug("function call: rename_on_db_folder")
    current, new = str(current), str(new)

    db_folder = path.split(DB_FILE)[0]
    path_to_current = path.join(db_folder, current)
    path_to_new = path.join(db_folder, new)

    try:
        rename(path_to_current, path_to_new)
        logger.info(f"{path_to_current} renomeado como {path_to_new}")
    except WindowsError:
        shutil.copy(path_to_current, path_to_new)
    except Exception as xpt:
        logger.error(f"Erro renomeando {path_to_current}: {xpt}", exc_info=1)
        if _raise:
            raise xpt


def check_sqlite(file: str, _raise: bool = False):
    """Checa se o full path para o arquivo `file` representa um banco de dados."""
    logger.debug("function call: check_sqlite")

    if not path.isfile(file):
        xpt = FileExistsError(f"Arquivo {file} invalido.")
        logger.error(str(xpt))
        if _raise:
            raise xpt

    con = sqlite3.connect(file)
    cursor = con.cursor()
    stmt = f"PRAGMA schema_version;"
    try:
        _ = cursor.execute(stmt).fetchone()
        if _ == (0,):
            raise sqlite3.DatabaseError()
        return True
    except sqlite3.DatabaseError:
        return False
    except Exception as xpt:
        logger.critical(f"Erro inesperado validando {file}", exc_info=1)
        if _raise:
            raise xpt
    finally:
        con.close()


def read_db_size(file_path: str = DB_FILE) -> int:
    logger.debug("function call: read_db_size")
    try:
        size = path.getsize(file_path)
        return size
    except Exception as xpt:
        logger.error(f"Erro lendo tamanho do arquivo: {str(xpt)}")
        return


####################
# LEITURAS
####################


def read_last_recorded_size(config_file: str = CONFIG_FILE):
    logger.debug("function call: read_last_recorded_size")
    config = configparser.ConfigParser()
    config.read(config_file)

    if "file_sizes" in config:
        return config["file_sizes"].getint("dbsize", fallback=None)
    return 0


####################
# ESCRITAS
####################


def write_current_size(current_size: int = read_db_size()) -> None:
    """Writes current database size to `backup.ini`"""
    logger.debug("function call: write_current_size")
    settings._write("data", "dbsize", current_size)


def write_add_backup_place(path: str):
    """Inclui o input `path` na opcao 'backup_places' em `backup.ini`"""
    logger.debug("function call: write_add_backup_place")
    settings._add_to_list("default", "backup_places", path)


def write_rm_backup_place(idx: int):
    """Retira o `idx`-esimo item da lista 'backup_places' em `backup.ini`"""
    logger.debug("function call: write_rm_backup_place")
    settings._rm_from_list("default", "backup_places", idx=idx)


def load(file: str, _raise: bool = False) -> None:
    """
    Checa se `file` se trata de um banco de dados SQLite valido, e entao o
    carrega como o banco de dados atual no Cashd.

    Se um banco de dados ja estiver presente, vai renomea-lo para um nome
    nao usado pelo Cashd nem por outros arquivos na pasta e o mantera no
    diretorio.
    """
    logger.debug("function call: load")
    db_is_present = path.isfile(DB_FILE)
    file_is_valid = check_sqlite(file)

    if not file_is_valid:
        msg = f"Impossivel carregar arquivo nao SQLite {file}"
        logger.error(msg)
        if _raise:
            raise OSError(msg)

    if db_is_present:
        now = datetime.now()
        dbfilename = path.split(DB_FILE)[1]
        stashfilename = f"stashed{now}.db".replace(":", "-")
        rename_on_db_folder(dbfilename, stashfilename)

    try:
        shutil.copyfile(file, DB_FILE)
    except shutil.SameFileError:
        pass


def run(force: bool = False, _raise: bool = False) -> None:
    """
    Vai fazer a copia do arquivo de banco de dados para a pasta local de backup
    e para as pastas listadas na opcao 'backup_places' em `backup.ini`.

    Usar `force = False` so vai fazer uma copia se o arquivo aumentou de
    tamanho, comparado com o registrado em 'file_sizes'.
    """
    backup_places = settings._read("default", "backup_places", convert_to="list")
    error_was_raised = False

    if not force:
        current_size = read_db_size()
        previous_size = settings.read_dbsize()
        if not previous_size:
            previous_size = 0
        if current_size <= previous_size:
            return
    settings.write_dbsize(current_size)

    try:
        backup_places = [i for i in [BACKUP_PATH] + backup_places if i != ""]
        for place in backup_places:
            try:
                if path.exists(place):
                    copy_file(DB_FILE, place, _raise=_raise)
                else:
                    raise NotADirectoryError(f"{place} nao existe")
            except Exception as xpt:
                logger.error(f"Nao foi possivel salvar em '{place}': {xpt}", exc_info=1)
                if _raise:
                    error_was_raised = True
    except Exception as xpt:
        logger.error(f"Erro inesperado durante o backup: {xpt}", exc_info=1)
    finally:
        if error_was_raised:
            raise NotADirectoryError(
                f"Erro em alguma etapa do backup, verifique o log: {LOG_FILE}"
            )



import plotly.graph_objects as pg
import pandas as pd
from datetime import datetime
from typing import Literal

from cashd import db


CORES = ["#478eff", "gray"]


def _preprocessar_data(
    tbl: pd.DataFrame,
    periodo: Literal["mes", "sem", "dia"],
    date_col: str = "Data",
) -> pd.DataFrame:
    """
    Retorna `tbl` com a coluna de data `date_col` formatada corretamente
    de acordo com a periodicidade em `periodo`
    """
    if periodo == "sem":
        tbl[date_col] = tbl[date_col].apply(
            lambda x: datetime.strptime(x + "-0", "%Y-%W-%w")
        )
    tbl[date_col] = pd.to_datetime(tbl.Data)
    return tbl


def _gerar_layout(
    tbl: pd.DataFrame, periodo: Literal["mes", "sem", "dia"], date_col: str = "Data"
) -> pg.Layout:
    """
    Retorna um `plotly.graph_objects.Layout` gerado para o conjunto de dados `tbl`
    """
    datestr = "%B de %Y"
    if periodo == "dia":
        datestr = "%d de %B de %Y"
    elif periodo == "sem":
        datestr = "%Y, semana %W"

    return pg.Layout(
        margin=dict(l=0, r=0, t=0, b=0),
        template="plotly_white",
        showlegend=False,
        hovermode="x unified",
        xaxis=dict(
            tickmode="array",
            tickvals=[i for i in tbl[date_col]],
            ticktext=[i.strftime(datestr) for i in tbl[date_col]],
            showticklabels=False,
        ),
        yaxis_tickprefix="R$",
        yaxis_tickformat=" ",
    )


def _gerar_layout_vazio():
    """
    Retorna um `plotly.graph_objects.Layout` sem marcadores de eixos
    """
    return pg.Layout(
        margin=dict(l=0, r=0, t=0, b=0),
        template="none",
        showlegend=False,
        xaxis=dict(showgrid=False, zeroline=False, visible=False),
        yaxis=dict(showgrid=False, zeroline=False, visible=False),
    )


def mensagem(msg: str):
    layout = _gerar_layout_vazio()
    fig = pg.Figure(layout=layout)
    fig.add_annotation(
        x=0.5,
        y=0.5,
        xref="paper",
        yref="paper",
        text="Nenhum dado disponível",
        showarrow=False,
        font=dict(size=20),
    )
    return fig


def balancos(periodo, n):
    tbl = _preprocessar_data(
        tbl=db.saldos_transac_periodo(periodo=periodo, n=n), periodo=periodo
    )
    if tbl.shape[0] == 0:
        return mensagem("Sem dados para exibir")

    tbl["SomasDisplay"] = tbl["Somas"].apply(
        lambda x: f"{x:_.2f}".replace(".", ",").replace("_", " ")
    )
    tbl["AbatDisplay"] = tbl["Abatimentos"].apply(
        lambda x: f"{x:_.2f}".replace(".", ",").replace("_", " ")
    )

    layout = _gerar_layout(tbl, periodo)

    fig = pg.Figure(layout=layout)
    fig.add_trace(
        pg.Bar(
            x=tbl["Data"],
            y=tbl["Somas"],
            name="Somas",
            customdata=tbl[["SomasDisplay"]],
            hovertemplate="<b>R$ %{customdata[0]}</b>",
            offsetgroup=0,
            marker=dict(color=CORES[1]),
        )
    )
    fig.add_trace(
        pg.Bar(
            x=tbl["Data"],
            y=tbl["Abatimentos"],
            name="Abatimentos",
            customdata=tbl[["AbatDisplay"]],
            hovertemplate="<b>R$ %{customdata[0]}</b>",
            offsetgroup=0,
            marker=dict(color=CORES[0]),
        )
    )
    fig._config["displayModeBar"] = False
    return fig


def saldo_acum(periodo, n):
    tbl = _preprocessar_data(
        tbl=db.saldos_transac_periodo(periodo=periodo, n=n), periodo=periodo
    )
    if tbl.shape[0] == 0:
        return mensagem("Sem dados para exibir")

    tbl["SaldoAcum"] = (tbl["Somas"] + tbl["Abatimentos"]).cumsum()
    tbl["SaldoAcumDisplay"] = tbl["SaldoAcum"].apply(
        lambda x: f"{x:_.2f}".replace(".", ",").replace("_", " ")
    )

    layout = _gerar_layout(tbl, periodo)

    fig = pg.Figure(layout=layout)
    fig.add_trace(
        pg.Scatter(
            x=tbl["Data"],
            y=tbl["SaldoAcum"],
            name="Saldo",
            mode="lines+markers",
            customdata=tbl[["SaldoAcumDisplay"]],
            hovertemplate="<b>R$ %{customdata[0]}</b>",
            offsetgroup=0,
            marker=dict(color=CORES[0]),
        )
    )
    fig.update_xaxes(showgrid=False)
    fig._config["displayModeBar"] = False
    return fig



PG_TRANSAC = """
<|layout|columns=.68fr auto 1fr|class_name=header_container|

<|part|class_name=header_logo|
<|Cashd|text|height=30px|width=30px|>
|>

<|part|class_name=align_item_stretch|
<|{nav_transac_val}|toggle|lov={nav_transac_lov}|on_change={lambda s: s.elem_transac_form.update_content(s, nav_transac_val[0])}|>
|>

<|part|class_name=text_right|class_name=header_top_right_corner|
<|🗕|button|on_action=btn_mudar_minimizado|>
<|🗖|button|on_action=btn_mudar_maximizado|>
<|✖|button|on_action=btn_encerrar|>
|>

|>

<br /><br />

<|layout|columns=1 1|columns[mobile]=1

<|part|partial={elem_transac_sel}|>

<|part|partial={elem_transac_form}|>

|>
"""


ELEMENTO_FORM = """
<|
__Data__*
<|{display_tr_data}|date|not editable|class_name=invisible|format=e|>

<|{display_tr_data}|date|>

__Valor__*: R$ <|{display_tr_valor}|text|>

<|{form_transac.Valor}|input|on_change=chg_transac_valor|change_delay=0|>

_(*) Obrigatório_
|>

<br />

<|Inserir|button|class_name=plain|on_action=btn_inserir_transac|>
"""


ELEMENTO_HIST = """
<|Excluir uma transação|button|on_action=btn_mostrar_dialogo_selec_transac|>
 
Saldo devedor atual: **R$ <|{SLC_USUARIO_SALDO}|>**

<|{df_transac}|table|paginated|height=300px|>

<|{mostra_selec_transac}|dialog|title=Qual transação será removida?|width=80%|partial={dial_selec_transac}|on_action=chg_dialog_selec_transac|page_id=selecionar_transacao|labels=Cancelar;Continuar|>

<|{mostra_confirma_transac}|dialog|title=Confirma remoção desta transação?|width=80%|partial={dial_transac_confirmar}|on_action=chg_dialog_confirma_transac|page_id=confirma_remover_transacao|labels=Voltar;Confirmar|>
"""


ELEMENTO_SELEC_CONTA = """
__Cliente__: _<|{nome_cliente_selec}|text|>_

__Local__: _<|{SLC_USUARIO_LOCAL}|text|>_

__Saldo devedor__: R$ <|{SLC_USUARIO_SALDO}|>

<|{SLC_USUARIO}|selector|lov={NOMES_USUARIOS}|filter|propagate|height=300px|width=450px|on_change=chg_cliente_selecionado|>
"""



from cashd import db, backup, plot, prefs
from cashd.pages import transac, contas, analise, configs, dialogo

from taipy.gui import Gui, notify, State, navigate, Icon, builder
from datetime import datetime
import tkinter as tk
from tkinter import filedialog
from tkinter.filedialog import askopenfilename
from pyshortcuts import make_shortcut
from os import path
import pandas as pd
import threading
import webview
import socket
import sys


PYTHON_PATH = path.dirname(sys.executable)

####################
# BOTOES
####################


def btn_mostrar_dialogo(state: State, id: str, payload: dict, show: str):
    show_dialogs = {
        "confirma_conta": "mostra_confirma_conta",
        "confirma_transac": "mostra_confirma_transac",
        "selec_cliente": "mostra_selec_cliente",
        "edita_cliente": "mostra_form_editar_cliente",
        "selec_transac": "mostra_selec_transac",
    }
    for dialog in show_dialogs.values():
        state.assign(dialog, False)
    state.assign(show_dialogs[show], True)


def btn_mostrar_dialogo_selec_cliente(state: State, id: str, payload: dict):
    btn_mostrar_dialogo(state, id, payload, show="selec_cliente")


def btn_mostrar_dialogo_edita_cliente(state: State, id: str, payload: dict):
    btn_mostrar_dialogo(state, id, payload, show="edita_cliente")


def btn_mostrar_dialogo_selec_transac(state: State, id: str, payload: dict):
    state.assign(
        "TRANSACS_USUARIO",
        db.listar_transac_cliente(state.SLC_USUARIO[0], para_mostrar=False),
    )
    btn_mostrar_dialogo(state, id, payload, "selec_transac")


def btn_atualizar_listagem(state: State):
    with db.DB_ENGINE.connect() as conn, conn.begin():
        state.df_clientes = pd.read_sql_query("SELECT * FROM clientes", con=conn)


def btn_gerar_main_plot(state: State | None = None):
    """
    Se `state=None` retorna um `plotly.graph_objects.Figure`, caso contrario, atualiza o valor
    de `'main_plot'`."""

    if state:
        p = state.dropdown_periodo_val[0]
        n = state.slider_val[0]
        tipo = state.dropdown_tipo_val
    else:
        p = dropdown_periodo_val[0]
        n = slider_val[0]
        tipo = dropdown_tipo_val

    if tipo == "Saldo Acumulado":
        fig = plot.saldo_acum(periodo=p, n=n)
    else:
        fig = plot.balancos(periodo=p, n=n)

    if state:
        state.assign("main_plot", fig)
        state.refresh("main_plot")
        return
    return


def btn_atualizar_df_ult_transac(state: State):
    try:
        updated_tbl = db.ultimas_transac_displ()
        state.assign("df_ult_transac", updated_tbl)
    except Exception as xpt:
        notify(state, "error", f"Erro inesperado atualizando a tabela: {str(xpt)}")


def btn_atualizar_locais_de_backup(state: State | None = None):
    """
    Se `state=None` retorna um `pd.DataFrame`, caso contrario, atualiza o valor
    de `'df_locais_de_backup'`."""
    locais_de_backup = backup.settings.read_backup_places()
    df = pd.DataFrame(
        {"Id": range(len(locais_de_backup)), "Locais de backup": locais_de_backup}
    )

    if state:
        state.assign("df_locais_de_backup", df)
        state.refresh("df_locais_de_backup")
        return
    return df


def btn_fazer_backups(state: State):
    try:
        backup.run(force=True, _raise=True)
        notify(state, "success", "Backup concluído!")
    except Exception as xpt:
        notify(state, "error", str(xpt))


def btn_add_local_de_backup(state: State):
    root = tk.Tk()
    root.withdraw()
    folder = filedialog.askdirectory()
    backup.prefs_.add_backup_place(folder)
    btn_atualizar_locais_de_backup(state)


def btn_rm_local_de_backup(state: State, var_name, payload):
    idx = int(payload["index"])
    backup.prefs_.rm_backup_place(idx)
    btn_atualizar_locais_de_backup(state)


def btn_carregar_backup(state: State):
    filename = askopenfilename()
    try:
        backup.load(file=filename, _raise=True)
        notify(state, "success", "Dados carregados com sucesso")
    except OSError:
        notify(state, "error", "Arquivo selecionado não é um banco de dados SQLite")
    except Exception as xpt:
        notify(state, "error", f"Erro inesperado carregando arquivo: {xpt}")


def btn_criar_atalho(state: State):
    ico_file = path.join(backup.SCRIPT_PATH, "assets", "ICO_LogoIcone.ico")
    startup_script = path.join(PYTHON_PATH, "cashd.exe")
    if not path.isfile(startup_script):
        startup_script = path.join(PYTHON_PATH, "Scripts", "cashd.exe")

    make_shortcut(
        script=startup_script,
        icon=ico_file,
        name="Cashd",
        description="Registre seu fluxo de caixa rapidamente e tenha total controle dos seus dados!",
        terminal=False,
        desktop=True,
        startmenu=True,
    )
    notify(state, "success", "Atalho criado com sucesso!")


def btn_inserir_transac(state: State):
    carregar_lista_transac(state)
    try:
        state.form_transac.DataTransac = state.display_tr_data
        nova_transac: dict = state.form_transac.despejar()
        state.form_transac.Valor = ""
        agora = datetime.now()

        db.adicionar_transac(db.tbl_transacoes(CarimboTempo=agora, **nova_transac))
        notify(state, "success", f"Nova transação adicionada!")

        id_selecionado = int(nova_transac["IdCliente"])
        state.assign("SLC_USUARIO", state.NOMES_USUARIOS[id_selecionado - 1])
        state.assign("display_tr_valor", "0,00")
        state.refresh("form_transac")
    except Exception as msg_erro:
        notify(state, "error", str(msg_erro))


def btn_inserir_cliente(state: State):
    try:
        novo_cliente: db.FormContas = state.form_contas.despejar()
        state.form_contas.__init__()
        db.adicionar_cliente(db.tbl_clientes(**novo_cliente))

        nome_completo = f"{novo_cliente['PrimeiroNome']} {novo_cliente['Sobrenome']}"
        notify(state, "success", message=f"Novo cliente adicionado!\n{nome_completo}")
        state.refresh("form_contas")
        state.NOMES_USUARIOS = sel_listar_clientes()
    except Exception as msg_erro:
        notify(state, "error", str(msg_erro))


def btn_chg_prefs_main_state(state: State, val: str):
    try:
        prefs.settings.write_main_state(val)
        state.form_contas.Estado = val
        state.refresh("form_contas")
        notify(state, "success", f"Estado preferido atualizado para {val}")
    except Exception as xpt:
        notify(state, "error", f"Erro inesperado: {str(xpt)}")


def btn_chg_prefs_main_city(state: State, val: str):
    try:
        val = val.title()
        prefs.settings.write_main_city(val)
        state.input_cidade_val = val
        state.form_contas.Cidade = val
        state.refresh("form_contas")
        notify(state, "success", f"Cidade preferida atualizada para {val}")
    except Exception as xpt:
        notify(state, "error", f"Erro inesperado: {str(xpt)}")


def btn_chg_max_ultimas_transacs(state: State, val: int):
    try:
        val = int(val)
        prefs.settings.write_last_transacs_limit(val)
        btn_atualizar_df_ult_transac(state)
        notify(
            state,
            "success",
            f"Limite de entradas em 'Últimas transações' atualizado para {val}",
        )
    except Exception as xpt:
        notify(state, "error", f"Erro inesperado: {str(xpt)}")


def btn_chg_max_highest_balances(state: State, val: int):
    try:
        val = int(val)
        prefs.settings.write_highest_balaces_limit(val)
        state.df_maiores_saldos = db.rank_maiores_saldos(val)
        notify(
            state,
            "success",
            f"Limite de entradas em 'Maiores saldos' atualizado para {val}",
        )
    except Exception as xpt:
        notify(state, "error", f"Erro inesperado: {str(xpt)}")


def tggl_backup_on_exit(state: State | None = None):
    if not state:
        return backup.settings.read_backup_on_exit()
    val = state.toggle_backup_on_exit
    backup.settings.write_backup_on_exit(val)


def btn_encerrar():
    try:
        backup.run(force=False, _raise=False)
        window.destroy()
    except NameError:
        pass
    finally:
        raise KeyboardInterrupt("Encerrando...")


def btn_mudar_maximizado():
    window.toggle_fullscreen()


def btn_mudar_minimizado():
    window.minimize()


####################
# UTILS
####################


def carregar_lista_transac(state: State):
    elems = db.listar_transac_cliente(state.SLC_USUARIO[0])
    state.df_transac = elems["df"]
    state.SLC_USUARIO_SALDO = elems["saldo"]
    state.SLC_USUARIO_LOCAL = elems["local"]
    state.refresh("df_transac")
    state.refresh("SLC_USUARIO_SALDO")


def sel_listar_clientes():
    clientes = db.listar_clientes()
    return [(str(i["id"]), i["nome"]) for i in clientes]


def menu_lateral(state, action, info):
    page = info["args"][0]
    navigate(state, to=page)


####################
# ON ACTION
####################


def chg_dialog_selec_cliente_conta(state: State, id: str, payload: dict):
    with state as s:
        if payload["args"][0] < 1:
            s.assign("mostra_selec_cliente", False)

        if payload["args"][0] == 1:
            if s.SLC_USUARIO[0] == "0":
                notify(s, "error", "Nenhuma conta foi selecionada")
            else:
                cliente_selec = db.cliente_por_id(s.SLC_USUARIO[0])
                s.form_conta_selec.carregar_valores(cliente_selec)
                s.refresh("form_conta_selec")
                s.assign("mostra_selec_cliente", False)
                s.assign("mostra_form_editar_cliente", True)


def chg_dialog_editar_cliente(state: State, id: str, payload: dict):
    with state as s:
        if payload["args"][0] == -1:
            s.assign("mostra_selec_cliente", False)

        if payload["args"][0] == 0:
            s.assign("mostra_form_editar_cliente", False)
            s.assign("mostra_selec_cliente", True)

        if payload["args"][0] == 1:
            s.assign("mostra_form_editar_cliente", False)
            s.assign("mostra_confirma_conta", True)


def chg_dialog_confirma_cliente(state: State, id: str, payload: dict):
    with state as s:
        if payload["args"][0] == 1:
            try:
                db.atualizar_cliente(state.SLC_USUARIO[0], state.form_conta_selec)
                state.NOMES_USUARIOS = sel_listar_clientes()
                notify(s, "success", "Cadastro atualizado com sucesso!")
            except Exception as xpt:
                notify(s, "error", f"Erro ao atualizar cadastro: {str(xpt)}")
                s.assign("mostra_confirma_conta", False)

        s.assign("mostra_confirma_conta", False)


def chg_dialog_selec_transac(state: State, id: str, payload: dict):
    with state as s:
        if payload["args"][0] < 1:
            s.assign("mostra_selec_transac", False)

        if payload["args"][0] == 1:
            if s.SLC_TRANSAC == "0":
                notify(s, "error", "Nenhuma transação foi selecionada")
            elif not db.id_transac_pertence_a_cliente(
                s.SLC_TRANSAC[0], s.SLC_USUARIO[0]
            ):
                notify(s, "error", "Selecione uma transação antes de continuar")
                return
            else:
                transac_selec = db.transac_por_id(s.SLC_TRANSAC[0])
                s.form_transac_selec.carregar_valores(transac_selec)
                s.refresh("form_transac_selec")
                s.assign("mostra_confirma_transac", True)
        s.assign("mostra_selec_transac", False)


def chg_dialog_confirma_transac(state: State, id: str, payload: dict):
    with state as s:
        s.assign("mostra_confirma_transac", False)

        if payload["args"][0] == 0:
            s.assign("mostra_selec_transac", True)

        if payload["args"][0] == 1:
            db.remover_transac(s.SLC_TRANSAC[0])
            notify(s, "success", "Transação removida.")
            s.assign("SLC_TRANSAC", "0")


def chg_transac_valor(state: State) -> None:
    state.display_tr_valor = db.fmt_moeda(state.form_transac.Valor, para_mostrar=True)
    state.refresh("form_transac")
    return


def chg_cliente_selecionado(state: State) -> None:
    carregar_lista_transac(state=state)
    state.form_transac.IdCliente = int(state.SLC_USUARIO[0])
    state.nome_cliente_selec = state.SLC_USUARIO[1]
    state.refresh("form_transac")


####################
# VALORES INICIAIS
####################

# visibilidade de dialogos
mostra_selec_cliente = False
mostra_selec_transac = False
mostra_form_editar_cliente = False
mostra_confirma_conta = False
mostra_confirma_transac = False

# controles dos graficos
slider_elems = list(range(10, 51)) + [None]
slider_lov = [(i, str(i)) if i is not None else (i, "Tudo") for i in slider_elems]
slider_val = slider_lov[0]

dropdown_periodo_lov = [("mes", "Mensal"), ("sem", "Semanal"), ("dia", "Diário")]
dropdown_periodo_val = dropdown_periodo_lov[0]

dropdown_tipo_lov = [
    "Balanço",
    "Saldo Acumulado",
]
dropdown_tipo_val = dropdown_tipo_lov[0]

dropdown_uf_lov = [
    "AC",
    "AL",
    "AP",
    "AM",
    "BA",
    "CE",
    "DF",
    "ES",
    "GO",
    "MA",
    "MT",
    "MS",
    "MG",
    "PA",
    "PB",
    "PE",
    "PI",
    "PR",
    "RJ",
    "RN",
    "RO",
    "RR",
    "RS",
    "SC",
    "SE",
    "SP",
    "TO",
]
dropdown_uf_val = prefs.settings.read_main_state()

main_plot = btn_gerar_main_plot()

# listagem de clientes
with db.DB_ENGINE.connect() as conn, conn.begin():
    df_clientes = pd.read_sql_query("SELECT * FROM clientes", con=conn)

# valor inicial dos campos "Valor" e "Data" no menu "Adicionar Transacao"
display_tr_valor = "0,00"
display_tr_data = datetime.now()

# valor inicial do seletor de conta global
NOMES_USUARIOS = sel_listar_clientes()
if len(NOMES_USUARIOS) > 0:
    SLC_USUARIO = NOMES_USUARIOS[0]
else:
    SLC_USUARIO = "0"

# formularios
form_contas = db.FormContas()
form_transac = db.FormTransac(IdCliente=SLC_USUARIO[0])
form_conta_selec = db.FormContas()
form_transac_selec = db.FormTransac(IdCliente=SLC_USUARIO[0])


# nome do cliente selecionado
nome_cliente_selec = ""

# valor inicial do seletor de transacao global
TRANSACS_USUARIO = db.listar_transac_cliente(SLC_USUARIO[0], para_mostrar=False)
if len(TRANSACS_USUARIO) > 0:
    SLC_TRANSAC = TRANSACS_USUARIO[0]
else:
    SLC_TRANSAC = "0"

# define se a webview vai iniciar em tela cheia
maximizado = False

# valor inicial da tabela de transacoes do usuario selecionado em SLC_USUARIO
df_ult_transac = db.ultimas_transac_displ()

df_maiores_saldos = db.rank_maiores_saldos()

# valor inicial do saldo do usuario selecionado em SLC_USUARIO
init_meta_cliente = db.listar_transac_cliente(SLC_USUARIO[0])

df_transac = init_meta_cliente["df"]
SLC_USUARIO_SALDO = init_meta_cliente["saldo"]
SLC_USUARIO_LOCAL = init_meta_cliente["local"]

# valor inicial da lista de locais de backup
df_locais_de_backup = btn_atualizar_locais_de_backup()

# valor inicial do campo "cidade preferida"
input_cidade_val = prefs.settings.read_main_city()

# valor inicial da configuracao Limite de linhas na tabela "Últimas transações"
input_quant_max_ultimas_transacs = prefs.settings.read_last_transacs_limit()
#                    " " "                       na tabela "Maiores saldos"
input_quant_max_highest_balances = prefs.settings.read_highest_balaces_limit()

# valor inicial do toggle "backup ao sair"
toggle_backup_on_exit = tggl_backup_on_exit()

# dados de entradas e abatimentos
df_entradas_abatimentos = db.saldos_transac_periodo()
layout_df_entradas_abatimentos = {
    "x": "Data",
    "y[1]": "Somas",
    "y[2]": "Abatimentos",
    "layout": {
        "barmode": "overlay",
        "barcornerradius": "20%",
        "hovermode": "x unified",
        "hovertemplate": "<b>Total</b>: R$ %{y:.2f}",
    },
}
config_df_entradas_abatimentos = {"displaymodebar": False}


RAIZ = """
<|menu|label=Menu|width=200px|lov={("transacoes", Icon("assets/SVG_TransacaoBranco.svg", "Transações")), ("clientes", Icon("assets/SVG_ContasBranco.svg", "Clientes")), ("analise", Icon("assets/SVG_DadosBranco.svg", "Estatísticas")), ("configs", Icon("assets/SVG_ConfiguracaoBranco.svg", "Configurações"))}|on_action=menu_lateral|>
"""


paginas = {
    "/": RAIZ,
    "transacoes": transac.PG_TRANSAC,
    "clientes": contas.PG_CONTAS,
    "analise": analise.PG_ANALISE,
    "configs": configs.PG_CONFIG,
}

app = Gui(pages=paginas, css_file="__main__.css")

elem_transac_sel = Gui.add_partial(app, transac.ELEMENTO_SELEC_CONTA)
elem_transac_form = Gui.add_partial(app, transac.ELEMENTO_FORM)
elem_conta = Gui.add_partial(app, contas.ELEMENTO_FORM)
elem_config = Gui.add_partial(app, configs.ELEMENTO_PREFS)
elem_analise = Gui.add_partial(app, analise.ELEM_HIST)

dial_selec_cliente = Gui.add_partial(app, dialogo.SELECIONAR_CLIENTE_ETAPA)
dial_selec_transac = Gui.add_partial(app, dialogo.SELECIONAR_TRANSAC_ETAPA)
dial_form_editar_cliente = Gui.add_partial(app, dialogo.FORM_EDITAR_CLIENTE)
dial_transac_confirmar = Gui.add_partial(app, dialogo.CONFIRMAR_TRANSAC)
dial_conta_confirmar = Gui.add_partial(app, dialogo.CONFIRMAR_CONTA)

### menus de navegacao ###
# transacoes
nav_transac_lov = [
    (transac.ELEMENTO_FORM, "Adicionar transação"),
    (transac.ELEMENTO_HIST, "Ver histórico"),
]
nav_transac_val = nav_transac_lov[0]
# contas
nav_conta_lov = [
    (contas.ELEMENTO_FORM, "Criar conta"),
    (contas.ELEMENTO_REGS, "Contas registradas"),
]
nav_conta_val = nav_conta_lov[0]
# estatisticas
nav_analise_lov = [(analise.ELEM_HIST, "Histórico"), (analise.ELEM_PLOT, "Gráficos")]
nav_analise_val = nav_analise_lov[0]
# configs
nav_config_lov = [
    (configs.ELEMENTO_PREFS, "Preferências"),
    (configs.ELEMENTO_BACKUP, "Backup"),
    (configs.ELEMENTO_ATALHO, "Outros"),
]
nav_config_val = nav_config_lov[0]


def porta_aberta() -> int:
    port = 5000
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        for i in range(51):
            if s.connect_ex(("localhost", port + i)) != 0:
                return port + i
        return port + 50


port = porta_aberta()


def start_cashd(with_webview: bool = False):
    def run_taipy_gui():
        app.run(
            title="Cashd",
            run_browser=not with_webview,
            dark_mode=False,
            stylekit={
                "color_primary": "#478eff",
                "color_background_light": "#ffffff",
            },
            run_server=True,
            port=port,
            favicon="assets/PNG_LogoFavicon.png",
            watermark="",
        )

    if with_webview:
        taipy_thread = threading.Thread(target=run_taipy_gui)
        taipy_thread.start()

        global window
        window = webview.create_window(
            title="Cashd",
            url=f"http://localhost:{port}",
            frameless=True,
            maximized=maximizado,
            easy_drag=False,
            min_size=(900, 600),
        )

        webview.start()

    else:
        run_taipy_gui()



PG_CONTAS = """
<|layout|columns=.68fr auto 1fr|class_name=header_container|

<|part|class_name=header_logo|
<|Cashd|text|height=30px|width=30px|>
|>

<|part|class_name=align_item_stretch|
<|{nav_conta_val}|toggle|lov={nav_conta_lov}|on_change={lambda s: s.elem_conta.update_content(s, nav_conta_val[0])}|>
|>

<|part|class_name=text_right|class_name=header_top_right_corner|
<|🗕|button|on_action=btn_mudar_minimizado|>
<|🗖|button|on_action=btn_mudar_maximizado|>
<|✖|button|on_action=btn_encerrar|>
|>

|>

<br />

<|part|partial={elem_conta}|class_name=container|>
"""


ELEMENTO_FORM = """
<|layout|columns=1 1|columns[mobile]=1 1|class_name=container
__Primeiro Nome__*

__Sobrenome__*

<|{form_contas.PrimeiroNome}|input|>

<|{form_contas.Sobrenome}|input|>

__Apelido__

__Telefone__

<|{form_contas.Apelido}|input|>

<|{form_contas.Telefone}|input|label=DDD + 9 dígitos|>

__Endereço__

__Bairro__

<|{form_contas.Endereco}|input|>

<|{form_contas.Bairro}|input|>

__Cidade__*

__Estado__*

<|{form_contas.Cidade}|input|>

<|{form_contas.Estado}|selector|lov={dropdown_uf_lov}|dropdown|>

_(*) Obrigatório_

<|Inserir|button|class_name=plain|on_action=btn_inserir_cliente|>
<|{str(form_contas)}|text|>

|>
"""


ELEMENTO_REGS = """
<|Atualizar listagem|button|class_name=plain|on_action={btn_atualizar_listagem}|> 
<|Editar uma conta|button|on_action={btn_mostrar_dialogo_selec_cliente}|>

<|{df_clientes}|table|paginated|filter|page_size=6|page_size_options={[12,24,36]}|height=380px|>

<|{mostra_selec_cliente}|dialog|title=Selecione o cliente que será editado|width=80%|partial={dial_selec_cliente}|on_action=chg_dialog_selec_cliente_conta|page_id=selecionar_conta|labels=Fechar;Continuar|>

<|{mostra_form_editar_cliente}|dialog|title=Editando...|width=80%|partial={dial_form_editar_cliente}|on_action=chg_dialog_editar_cliente|page_id=editar_conta|labels=Voltar;Continuar|>

<|{mostra_confirma_conta}|dialog|title=Tem certeza?|width=80%|partial={dial_conta_confirmar}|on_action=chg_dialog_confirma_cliente|page_id=confirma_editar_conta|labels=Cancelar;Confirmar|>
"""



import pandas as pd
from cashd import prefs
from datetime import datetime, date
from sqlalchemy import (
    ForeignKey,
    Integer,
    create_engine,
    DateTime,
    Date,
    insert,
    update,
    select,
    delete,
    case,
    text,
)
from sqlalchemy.orm import (
    Mapped,
    mapped_column,
    DeclarativeBase,
    relationship,
    Session,
    Query,
)
from copy import deepcopy
from typing import List, Literal
from time import sleep
from platform import system
import os
from os import path
import phonenumbers


####################
# VARIAVEIS DO BANCO DE DADOS
####################

SCRIPT_FOLDER = path.dirname(path.realpath(__file__))
DATA_FOLDER = SCRIPT_FOLDER + "\\data"
os.makedirs(DATA_FOLDER, exist_ok=True)
DB_ENGINE = create_engine(f"sqlite:///{DATA_FOLDER}\\database.db", echo=False)


class dec_base(DeclarativeBase):
    pass


class tbl_clientes(dec_base):
    __tablename__ = "clientes"
    SaldoTransacoes: Mapped[List["tbl_transacoes"]] = relationship()

    Id: Mapped[int] = mapped_column(primary_key=True)
    PrimeiroNome: Mapped[str] = mapped_column()
    Sobrenome: Mapped[str] = mapped_column()
    Apelido: Mapped[str] = mapped_column()
    Telefone: Mapped[str] = mapped_column()
    Cidade: Mapped[str] = mapped_column()
    Bairro: Mapped[str] = mapped_column()
    Endereco: Mapped[str] = mapped_column()
    Estado: Mapped[str] = mapped_column()


class tbl_transacoes(dec_base):
    __tablename__ = "transacoes"
    NomeCliente: Mapped["tbl_clientes"] = relationship(back_populates="SaldoTransacoes")

    Id: Mapped[int] = mapped_column(primary_key=True)
    IdCliente: Mapped[int] = mapped_column(ForeignKey("clientes.Id"))
    CarimboTempo: Mapped[datetime] = mapped_column(DateTime(timezone=True))
    DataTransac: Mapped[date] = mapped_column(Date())
    Valor: Mapped[int] = mapped_column()  # valor em centavos


dec_base.metadata.create_all(DB_ENGINE)

SYS_NAME = system()
try:
    if SYS_NAME == "Windows":
        WORK_DIR = path.expanduser("~\\Appdata\\Roaming\\Cashd")

    elif SYS_NAME == "Darwin":
        WORK_DIR = path.expanduser("~/Library/Preferences/Cashd")

    else:
        WORK_DIR = path.expanduser("~/.local/Share/Cashd")

    os.makedirs(WORK_DIR, exist_ok=True)
    os.chdir(WORK_DIR)

except Exception as manage_dir_error:
    sleep_time = 30
    print("ERRO FATAL:\n", manage_dir_error, "\nFechando em 30 segundos...", sep="")
    sleep(30)
    quit()


####################
# FORMATACAO
####################


class ErroCampoVazio(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class ErroDeFormatacao(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


def fmt_moeda(inp: str, para_mostrar: bool = False) -> str:
    """Formata números inteiros (em centavos) para R$.

    `para_mostrar == False`:
    - 100000 -> 100000
    - -5254 -> -5254
    - 0 -> ErroDeFormatacao
    - 58/66 -> ErroDeFormatacao

    `para_mostrar == True`:
    - 100000 ->  `'1 000,00'`
    - -5254 -> `'-52,54'`
    - 0 -> `'0,00'`
    - 58/66 -> `'0,00'`
    """
    if inp is str:
        inp = inp.replace(",", "")

    if para_mostrar:
        try:
            return f"{int(inp)/100:_.2f}".replace("_", " ").replace(".", ",")

        except ValueError:
            return "0,00"

    else:
        try:
            out = int(inp)
            if out == 0:
                raise ErroDeFormatacao("Valor não pode ser zero")

        except ValueError:
            raise ErroDeFormatacao("Valor deve conter apenas números")

    return out


def fmt_telefone(inp: str, para_mostrar: bool = False) -> str:
    """"""
    try:
        return phonenumbers.format_number(
            numobj=phonenumbers.parse(inp, "BR"),
            num_format=phonenumbers.PhoneNumberFormat.NATIONAL,
        )

    except phonenumbers.NumberParseException:
        if para_mostrar:
            return "Número inválido"
        else:
            raise ErroDeFormatacao("Número de telefone inválido")


def fmt_nome_proprio(
    inp: str,
    permitido_vazio: bool = False,
) -> str:

    if not permitido_vazio and len(inp) == 0:
        raise ErroCampoVazio("Nome obrigatório não pode ficar vazio")

    return inp.title()


####################
# FORMULARIOS
####################


class NomeObrigatorio(str):
    def __init__(self) -> None:
        super().__init__()


class NomeOpcional(str):
    def __init__(self) -> None:
        super().__init__()


class NumeroTelefone(str):
    def __init__(self) -> None:
        super().__init__()


class NomeObrigatorioMaiusculo(str):
    def __init__(self) -> None:
        super().__init__()


class NumeroMoeda(int):
    def __init__(self) -> None:
        super().__init__()


class FormObj:
    def format_vars(self):
        _types = self.__init__.__annotations__

        for var in _types.keys():
            value = getattr(self, var)

            if _types[var] == NomeObrigatorio:
                try:
                    setattr(self, var, fmt_nome_proprio(value, False))
                except Exception as msg_erro:
                    raise ErroDeFormatacao(f"Erro em '{var}':\n{str(msg_erro)}")

            if _types[var] == NomeOpcional:
                try:
                    setattr(self, var, fmt_nome_proprio(value, True))
                except Exception as msg_erro:
                    raise ErroDeFormatacao(f"Erro em '{var}':\n{str(msg_erro)}")

            elif _types[var] == NumeroTelefone:
                try:
                    setattr(self, var, fmt_telefone(value))
                except Exception as msg_erro:
                    raise ErroDeFormatacao(f"Erro em '{var}':\n{str(msg_erro)}")

            elif _types[var] == NumeroMoeda:
                try:
                    setattr(self, var, fmt_moeda(value))
                except Exception as msg_erro:
                    raise ErroDeFormatacao(f"Erro em '{var}':\n{str(msg_erro)}")

            elif _types[var] == NomeObrigatorioMaiusculo:
                try:
                    setattr(self, var, fmt_nome_proprio(value, False).upper())
                except Exception as msg_erro:
                    raise ErroDeFormatacao(f"Erro em '{var}':\n{str(msg_erro)}")

    def despejar(self):
        self.format_vars()
        return deepcopy(vars(self))


class FormContas(FormObj):
    def __init__(
        self,
        PrimeiroNome: NomeObrigatorio = "",
        Sobrenome: NomeObrigatorio = "",
        Apelido: NomeOpcional = "",
        Telefone: NumeroTelefone = "81900000000",
        Cidade: NomeObrigatorio = prefs.settings.read_main_city(),
        Bairro: NomeOpcional = "",
        Endereco: NomeOpcional = "",
        Estado: NomeObrigatorioMaiusculo = prefs.settings.read_main_state(),
    ):

        self.PrimeiroNome = PrimeiroNome
        self.Sobrenome = Sobrenome
        self.Apelido = Apelido
        self.Telefone = Telefone
        self.Cidade = Cidade
        self.Bairro = Bairro
        self.Endereco = Endereco
        self.Estado = Estado

    def carregar_valores(self, tbl: tbl_clientes):
        self.PrimeiroNome = tbl.PrimeiroNome
        self.Sobrenome = tbl.Sobrenome
        self.Apelido = tbl.Apelido
        self.Telefone = tbl.Telefone
        self.Cidade = tbl.Cidade
        self.Bairro = tbl.Bairro
        self.Endereco = tbl.Endereco
        self.Estado = tbl.Estado

    def __str__(self):
        return ""

    def __repr__(self):
        return (
            "cashd.db.FormContas object. Fields:\n"
            + f"{self.PrimeiroNome=}\n"
            + f"{self.Sobrenome=}\n"
            + f"{self.Apelido=}\n"
            + f"{self.Telefone=}\n"
            + f"{self.Cidade=}\n"
            + f"{self.Bairro=}\n"
            + f"{self.Endereco=}\n"
            + f"{self.Estado=}"
        )


class FormTransac(FormObj):
    def __init__(
        self, IdCliente, DataTransac: date = date.today(), Valor: NumeroMoeda = ""
    ):

        self.IdCliente = IdCliente
        self.DataTransac = DataTransac
        self.Valor = Valor

    def carregar_valores(self, tbl: tbl_transacoes):
        self.IdCliente = tbl.IdCliente
        self.DataTransac = tbl.DataTransac
        self.Valor = fmt_moeda(tbl.Valor, True)

    def __repr__(self):
        return f"{self.IdCliente=}\n" + f"{self.DataTransac=}\n" + f"{self.Valor=}\n"


####################
# QUERRIES
####################


def listar_clientes() -> None:
    with Session(DB_ENGINE) as ses:
        stmt = select(tbl_clientes)
        res = ses.execute(stmt).scalars()

        output = []
        for r in res:
            linha = {"id": r.Id, "nome": f"{r.Id}, {r.PrimeiroNome} {r.Sobrenome}"}

            if r.Apelido != "":
                linha["nome"] = linha["nome"] + f" ({r.Apelido})"
            elif r.Apelido == "" and r.Bairro != "":
                linha["nome"] = linha["nome"] + f", {r.Bairro}"

            output.append(linha)

    return output


def listar_transac_cliente(Id: int, para_mostrar: bool = True) -> dict | list:
    Id = int(Id)

    stmt = select(
        tbl_transacoes.Id, tbl_transacoes.DataTransac, tbl_transacoes.Valor
    ).where(tbl_transacoes.IdCliente == Id)

    with Session(DB_ENGINE) as ses:
        res = ses.execute(stmt).all()

        if para_mostrar:
            df = pd.DataFrame([row[1:] for row in res], columns=["Data", "Valor R$"])
            df["Valor R$"] = list(map(lambda x: fmt_moeda(x, True), df["Valor R$"]))
            df = df.loc[::-1].sort_values("Data", axis=0, ascending=False)
            saldo = fmt_moeda(sum(r.Valor for r in res), para_mostrar=True)
            local = local_por_id(Id=Id)
            return {"df": df, "saldo": saldo, "local": local}

        return list(
            reversed(
                [
                    (
                        str(row[0]),
                        f"{row[1].strftime('%d/%m/%Y')} | {fmt_moeda(row[2], True)}",
                    )
                    for row in res
                ]
            )
        )


def adicionar_cliente(cliente: tbl_clientes) -> None:
    with Session(DB_ENGINE) as ses:
        ses.add(cliente)
        ses.commit()


def atualizar_cliente(Id: int, form: FormContas) -> None:
    Id = int(Id)
    table_values = form.despejar()

    with Session(DB_ENGINE) as ses:
        stmt = update(tbl_clientes).where(tbl_clientes.Id == Id).values(**table_values)
        ses.execute(stmt)
        ses.commit()


def adicionar_transac(transac: tbl_transacoes) -> None:
    with Session(DB_ENGINE) as ses:
        ses.add(transac)
        ses.commit()


def remover_transac(Id: int) -> None:
    with Session(DB_ENGINE) as ses:
        stmt = delete(tbl_transacoes).where(tbl_transacoes.Id == Id)
        ses.execute(stmt)
        ses.commit()


def cliente_por_id(Id: int) -> tbl_clientes:
    """retorna um objeto `db.tbl_clientes` com o id especificado."""
    Id = int(Id)
    with Session(DB_ENGINE) as ses:
        stmt = select(tbl_clientes).where(tbl_clientes.Id == Id)
        return ses.execute(stmt).first()[0]


def transac_por_id(Id: int) -> tbl_transacoes:
    """retorna um objeto `db.tbl_transacoes` com o id especificado."""
    Id = int(Id)
    with Session(DB_ENGINE) as ses:
        stmt = select(tbl_transacoes).where(tbl_transacoes.Id == Id)
        return ses.execute(stmt).first()[0]


def id_transac_pertence_a_cliente(IdTransac: int, IdCliente: int) -> bool:
    """Retorna `True` se a transacao pertence ao cliente, `False` em qualquer outro caso."""
    IdTransac, IdCliente = int(IdTransac), int(IdCliente)
    stmt = select(tbl_transacoes).filter_by(Id=IdTransac, IdCliente=IdCliente)
    with Session(DB_ENGINE) as ses:
        if ses.execute(stmt).first() is None:
            return False
        return True


def saldos_transac_periodo(
    periodo: Literal["mes", "sem", "dia"] = "mes", n: int | None = None
) -> pd.DataFrame:
    """
    Retorna um `pandas.DataFrame` com três colunas:
    - Data
    - Somas
    - Abatimentos

    `n` indica a quantidade de linhas que devem ser retornadas, começando pela
    última, se `n=None`, retornará todas as linhas.
    """
    formato = "%Y-%m"
    if periodo == "sem":
        formato = "%Y-%W"
    if periodo == "dia":
        formato = "%Y-%m-%d"

    stmt = f"""
    select
        STRFTIME('{formato}', "DataTransac") AS MesTransac,
        SUM(CASE WHEN Valor>0 THEN Valor ELSE 0 END)*1.0/100 AS Somas,
        SUM(CASE WHEN Valor<0 THEN Valor ELSE 0 END)*1.0/100 AS Abatimentos
    from transacoes
    group by STRFTIME('{formato}', "DataTransac");
    """

    with Session(DB_ENGINE) as ses:
        result = ses.execute(text(stmt)).all()
        tbl = pd.DataFrame(
            [res for res in result], columns=["Data", "Somas", "Abatimentos"]
        )
        if n:
            tbl = tbl.tail(n)
        return tbl


def ultimas_transac(n: int | None = prefs.settings.read_last_transacs_limit()):
    """
    Lista as ultimas `n` transacoes, comecando pela mais recente.
    Se `n=None`, retorna todas as transacoes.

    Nome das colunas:

    - `Data`
    - `Valor`
    - `Id do cliente`
    """

    stmt = f"""
    SELECT DataTransac, Valor, IdCliente
    FROM transacoes
    ORDER BY Id desc;
    """

    if n is not None:
        stmt = stmt.replace(";", f" LIMIT {n};")

    with Session(DB_ENGINE) as ses:
        result = ses.execute(text(stmt)).all()
        return pd.DataFrame(
            [res for res in result], columns=["Data", "Valor", "Id do cliente"]
        )


def nome_por_id(Id: int):
    c = cliente_por_id(Id=Id)
    if c.Apelido != "":
        return f"{c.PrimeiroNome} {c.Sobrenome} ({c.Apelido})"
    if c.Bairro != "":
        return f"{c.PrimeiroNome} {c.Sobrenome}, {c.Bairro}"
    return f"{c.PrimeiroNome} {c.Sobrenome}"


def local_por_id(Id: int) -> str:
    """Retorna um texto com informacoes de localizacao da conta selecionada"""
    try:
        c = cliente_por_id(Id=Id)
    except TypeError:
        # retorna string vazia se nao houverem cadastros
        return ""

    logradouro = ""
    if c.Endereco != "":
        logradouro = logradouro + f"{c.Endereco}, "
    if c.Bairro != "":
        logradouro = logradouro + f"{c.Bairro} - "
    return logradouro + f"{c.Cidade}/{c.Estado}"


def ultimas_transac_displ(n: int | None = prefs.settings.read_last_transacs_limit()):
    tbl = ultimas_transac(n=n)
    tbl["Cliente"] = pd.Series(map(nome_por_id, tbl["Id do cliente"]))
    tbl["Valor"] = tbl["Valor"].apply(lambda x: fmt_moeda(x, para_mostrar=True))
    tbl.drop("Id do cliente", axis="columns", inplace=True)
    return tbl


def rank_maiores_saldos(n=prefs.settings.read_highest_balaces_limit()):
    stmt = """
    SELECT IdCliente, SUM(Valor) AS Saldo
    FROM transacoes
    GROUP BY IdCliente
    ORDER BY Saldo DESC;
    """
    if n:
        stmt = stmt.replace(";", f" LIMIT {n};")

    with Session(DB_ENGINE) as ses:
        result = ses.execute(text(stmt)).all()

    tbl = pd.DataFrame([res for res in result], columns=["Cliente", "Saldo"])
    tbl["Cliente"] = pd.Series(map(nome_por_id, tbl["Cliente"]))
    tbl["Saldo"] = tbl["Saldo"].apply(lambda x: fmt_moeda(x, True))
    return tbl



from os import path, makedirs
from typing import Literal
import configparser
import logging


SCRIPT_PATH = path.split(path.realpath(__file__))[0]
PREFS_CONFIG_FILE = path.join(SCRIPT_PATH, "configs", "prefs.ini")
BACKUP_CONFIG_FILE = path.join(SCRIPT_PATH, "configs", "backup.ini")
LOG_FILE = path.join(SCRIPT_PATH, "logs", "prefs.log")
DB_FILE = path.join(SCRIPT_PATH, "data", "database.db")


class SettingsHandler:
    """
    Valores de configuração usados:

    ### prefs.ini

    `[default]`
    - last_transacs_limit: `int`
    - highest_balaces_limit: `int`
    - main_state: `str`
    - main_city: `str`

    ### backup.ini

    `[default]`
    - backup_on_exit: `bool`
    - backup_places: `list`

    `[data]`
    - dbsize: `int`
    """

    def __init__(self, configname):
        self.config_file = path.join(SCRIPT_PATH, "configs", f"{configname}.ini")
        self.log_file = path.join(SCRIPT_PATH, "logs", f"{configname}.log")

        # config parser
        self.conf = configparser.ConfigParser()
        self.conf.read(self.config_file, "utf-8")
        try:
            self.conf.add_section("default")
        except configparser.DuplicateSectionError:
            pass

        # logger
        self.logger = logging.getLogger(f"cashd.{__name__}")
        self.logger.setLevel(logging.DEBUG)
        self.logger.propagate = False

        log_fmt = logging.Formatter("%(asctime)s :: %(levelname)s %(message)s")
        log_handler = logging.FileHandler(LOG_FILE)
        log_handler.setLevel(logging.DEBUG)
        log_handler.setFormatter(log_fmt)
        self.logger.addHandler(log_handler)

        # create files, if not exist
        for file in [self.config_file, self.log_file]:
            makedirs(path.split(file)[0], exist_ok=True)
            if not path.isfile(file):
                with open(file=file, mode="a"):
                    pass

    def parse_list_from_config(self, string: str) -> list[str]:
        """
        Transforma uma config com multiplos itens uma uma lista de strings
        do python.
        """
        string = string.replace("[", "").replace("]", "")
        list_of_items = string.split(",")
        return [i.strip() for i in list_of_items if i.strip() != ""]

    def parse_list_to_config(self, list_: list[str]) -> str:
        """
        Transforma uma lista de strings do python em uma config (str) com mais
        de um item.
        """
        string_list = (
            str(list_).replace("[", "[\n\t").replace(", ", ",\n\t").replace("'", "")
        )
        return string_list.replace("\\\\", "\\")

    def _write(self, sect: str, key: str, val: str):
        """Escreve a combinacao de `key` e `val` na seção `sect`"""
        try:
            if not self.conf.has_section(sect):
                self.conf.add_section(sect)

            self.conf.set(sect, key, val)
            with open(self.config_file, "w") as newconfig:
                self.conf.write(newconfig)
            self.conf.read(self.config_file, "utf-8")
            self.logger.info(
                f"Valor atualizado em {self.config_file}: [{sect}] {key} = {val}"
            )

        except Exception as xpt:
            self.logger.error(f"Erro escrevendo [{sect}] {key}={val}: {str(xpt)}")
            raise xpt

    def _read(
        self,
        sect: str,
        key: str,
        convert_to: Literal[None, "bool", "int", "list"] = None,
    ):
        try:
            if not convert_to:
                return self.conf[sect][key]
            elif convert_to == "bool":
                return self.conf.getboolean(sect, key)
            elif convert_to == "int":
                return self.conf.getint(sect, key)
            elif convert_to == "list":
                return self.parse_list_from_config(self.conf.get(sect, key))

        except KeyError:
            return None
        except configparser.NoSectionError:
            return None
        except configparser.NoOptionError:
            return None

    def _add_to_list(self, sect: str, key: str, val: str):
        current_list = self.parse_list_from_config(self.conf[sect][key])
        new_list = set(current_list + [val])

        self.conf.set(sect, key, self.parse_list_to_config(new_list))
        with open(self.config_file, "w") as newconfig:
            self.conf.write(newconfig)
        self.conf.read(self.config_file, "utf-8")

    def _rm_from_list(self, sect: str, key: str, idx: int):
        """Retira o `idx`-esimo item da lista"""
        current_list = self.parse_list_from_config(self.conf[sect][key])
        n = len(current_list)

        if (idx + 1) > n:
            self.logger.error(f"{idx} fora dos limites, deve ser menor que {n}")

        _ = current_list.pop(idx)
        self.conf.set(sect, key, self.parse_list_to_config(current_list))
        with open(self.config_file, "w") as newconfig:
            self.conf.write(newconfig)
        self.conf.read(self.config_file, "utf-8")


class PreferencesHandler(SettingsHandler):
    def __init__(self, configname="prefs"):
        super().__init__(configname)

        # set defaults
        if self.read_last_transacs_limit() is None:
            self.write_last_transacs_limit(1000)

        if self.read_highest_balaces_limit() is None:
            self.write_highest_balaces_limit(10)

        if self.read_main_state() is None:
            self.write_main_state("AC")

        if self.read_main_city() is None:
            self.write_main_city("")

    def write_last_transacs_limit(self, val: int):
        """
        Controla um limite de transações a ser exibidas na tabela
        'Últimas transações'.
        """
        val = int(val)
        self._write("default", "last_transacs_limit", str(val))

    def write_highest_balaces_limit(self, val: str):
        """
        Controla um limite de contas a ser exibidas na tabela
        'Maiores saldos'.
        """
        val = int(val)
        self._write("default", "highest_balaces_limit", str(val))

    def write_main_state(self, val: str):
        """Controla o UF padrão exibido no formulário 'Criar conta'"""
        self._write("default", "main_state", val)

    def write_main_city(self, val: str):
        """Controla a cidade padrão exibida no formulário 'Criar conta'"""
        self._write("default", "main_city", val)

    def read_last_transacs_limit(self) -> int | None:
        return self._read("default", "last_transacs_limit", convert_to="int")

    def read_highest_balaces_limit(self) -> int | None:
        return self._read("default", "highest_balaces_limit", convert_to="int")

    def read_main_state(self) -> str | None:
        return self._read("default", "main_state")

    def read_main_city(self) -> str | None:
        return self._read("default", "main_city")


class BackupPrefsHandler(SettingsHandler):
    def __init__(self, configname="backup"):
        super().__init__(configname)

        if self.read_backup_places() is None:
            self._write("default", "backup_places", "[]")

        if self.read_backup_on_exit() is None:
            self._write("default", "backup_on_exit", "true")

    def read_backup_places(self) -> str | None:
        return self._read("default", "backup_places", convert_to="list")

    def read_backup_on_exit(self) -> bool | None:
        return self._read("default", "backup_on_exit", convert_to="bool")
    
    def read_dbsize(self) -> int | None:
        return self._read("data", "dbsize", convert_to="int")
    
    def write_dbsize(self, val: int) -> None:
        val = int(val)
        self._write("data", "dbsize", str(val))

    def write_backup_on_exit(self, val: bool) -> None:
        val = str(val).lower()
        self._write("default", "backup_on_exit", val)

    def add_backup_place(self, place: str) -> None:
        self._add_to_list("default", "backup_places", place)

    def rm_backup_place(self, idx: int) -> None:
        self._rm_from_list("default", "backup_places", idx)


###########################
# init and write defaults #
###########################

settings = PreferencesHandler()



PG_CONFIG = """
<|layout|columns=.68fr auto 1fr|persistent|class_name=header_container|

<|part|class_name=header_logo|
<|Cashd|text|height=30px|width=30px|>
|>

<|part|class_name=align_item_stretch|
<|{nav_config_val}|toggle|lov={nav_config_lov}|on_change={lambda s: s.elem_config.update_content(s, nav_config_val[0])}|>
|>

<|part|class_name=text_right|class_name=header_top_right_corner|
<|🗕|button|on_action=btn_mudar_minimizado|>
<|🗖|button|on_action=btn_mudar_maximizado|>
<|✖|button|on_action=btn_encerrar|>
|>

|>

<br />

<|part|partial={elem_config}|class_name=narrow_element|>

<br />
"""

ELEMENTO_BACKUP = """
# Locais de backup

Clique no símbolo de **+** abaixo para adicionar um **Local de backup**:

<br />

<|{df_locais_de_backup}|table|page_size=5|on_add={btn_add_local_de_backup}|on_delete={btn_rm_local_de_backup}|height=180px|>

# Ações

<|layout|columns=1fr 1fr|

<|part|
<|Carregar backup|button|on_action={btn_carregar_backup}|>

*_Não se preocupe, esta operação é reversível. Consulte a documentação._
|>

<|part|
<|Fazer backup|button|class_name=plain|on_action={btn_fazer_backups}|>

*_Backups serão salvos nos Locais de backup._
|>
|>
"""

ELEMENTO_ATALHO = """
# Atalhos

<|Adicionar atalho|button|on_action={btn_criar_atalho}|class_name=plain|>

*_Atalhos serão adicionados ao menu iniciar e à área de trabalho._

<br />

# Sessão

Executando em http://127.0.0.1:<|{port}|text|>
"""

ELEMENTO_PREFS = """
# Contas

__Valores padrão no formulário:__

<br />

<|layout|columns=1 1|columns[mobile]=1 1|class_name=container

<|{dropdown_uf_val}|selector|label=Estado padrão|lov={dropdown_uf_lov}|dropdown|on_change={lambda s: btn_chg_prefs_main_state(s, dropdown_uf_val)}|>

<|{input_cidade_val}|input|label=Cidade padrão|change_delay=1200|on_change={lambda s: btn_chg_prefs_main_city(s, input_cidade_val)}|>

|>

# Estatísticas

__Limite de linhas na tabela:__

<br />

<|layout|columns=1 1|columns[mobile]=1 1|class_name=container

<|{input_quant_max_ultimas_transacs}|number|label=Últimas transações [1000]|change_delay=1200|on_change={lambda s: btn_chg_max_ultimas_transacs(s, input_quant_max_ultimas_transacs)}|>

<|{input_quant_max_highest_balances}|number|label=Maiores saldos [10]|change_delay=1200|on_change={lambda s: btn_chg_max_highest_balances(s, input_quant_max_highest_balances)}|>

|>

"""
